"""Implementations of detection algorithms."""

from typing import Literal, Dict, Callable
from .utils import load_model
from .methods import log_likelihood, log_rank, likelihood_logrank_ratio, fast_detect_gpt
import torch
from transformers import AutoTokenizer, AutoModelForCausalLM

DETECTION_FUNCS: Dict[str, Callable[[torch.Tensor, torch.Tensor], float]] = {
    "loglikelihood": log_likelihood,
    "logrank": log_rank,
    "detectllm": likelihood_logrank_ratio,
    "fastdetectgpt": fast_detect_gpt,
}
THRESHOLDS: Dict[str, float] = {
    "loglikelihood": -1.8,
    "logrank": -0.8,
    "detectllm": 2.14,
    "fastdetectgpt": 1.9,
}

DetectionMethod = Literal["loglikelihood", "logrank", "detectllm", "fastdetectgpt"]


def detect_ai_text(
    text: str,
    method: DetectionMethod = "fastdetectgpt",
    threshold: float = None,
    detection_model: str = "Qwen/Qwen2.5-1.5B",
) -> int:
    """Detect if `text` is written by human or ai.

    Args:
        text (str): The text to check.
        method (DetectionMethod, optional), default='fastdetectgpt': Detection method to use, must be one of ['loglikelihood', 'logrank', 'detectllm', 'fastdetectgpt']
        threshold (float | None, optional), default=None: Decision threshold for `method` to use. If not provided, a default value will be used based on `method`.
        detection_model (str, optional), default=Qwen/Qwen2.5-1.5B: Huggingface Repo name for the model that `method` will use to generate logits.

    Returns:
        int: 0 if human generated 1 if machine generated.

    Raises:
        ValueError: If method is not one of ['loglikelihood', 'logrank', 'detectllm', 'fastdetectgpt'].
    """
    if not text:
        return 0

    device: str = "cuda" if torch.cuda.is_available() else "cpu"
    model: AutoModelForCausalLM
    tokenizer: AutoTokenizer
    model, tokenizer = load_model(detection_model)

    tokens: torch.Tensor = tokenizer(
        text,
        return_tensors="pt",
        padding=True,
        return_token_type_ids=False,
    ).to(device)

    if method not in DETECTION_FUNCS or method not in THRESHOLDS:
        raise ValueError(
            f"In detect_ai_text `method` must be one of ['loglikelihood', 'logrank', 'detectllm', 'fastdetectgpt'], but got {method}"
        )

    method_func: Callable[[torch.Tensor, torch.Tensor], float] = DETECTION_FUNCS[method]
    if threshold is None:
        threshold = THRESHOLDS[method]

    labels: torch.Tensor = tokens.input_ids[:, 1:]  # remove bos token
    with torch.no_grad():
        logits: torch.Tensor = model(**tokens).logits[
            :, :-1
        ]  # remove next token logits
    pred: float = method_func(labels, logits)

    return 0 if pred < threshold else 1



"""Detection functions."""

import torch
import torch.nn.functional as F
from .utils import validate_tensor_shapes


def log_likelihood(labels: torch.Tensor, logits: torch.Tensor) -> float:
    """Compute the loglikelihood of labels in logits.

    Args:
        labels (torch.Tensor): Ground truth labels of shape: (1, sequence_length).
        logits (torch.Tensor): Logits of shape: (1, sequence_length, vocab_size).

    Returns:
        float: The mean loglikelihood.

    Raises:
        ValueError: If the shapes of `labels` and `logits` are incompatible or batch size is > 1.
    """
    validate_tensor_shapes(labels, logits)

    logits: torch.Tensor = logits.view(-1, logits.shape[-1])
    labels: torch.Tensor = labels.view(-1)

    log_probs: torch.Tensor = F.log_softmax(logits, dim=-1)
    actual_token_probs: torch.Tensor = log_probs.gather(
        dim=-1, index=labels.unsqueeze(-1)
    ).squeeze(-1)
    return actual_token_probs.mean().item()


def log_rank(labels: torch.Tensor, logits: torch.Tensor) -> float:
    """Compute the negative average log rank of labels in logits.

    Args:
        labels (torch.Tensor): Ground truth labels of shape: (1, sequence_length).
        logits (torch.Tensor): Logits of shape: (1, sequence_length, vocab_size).

    Returns:
        float: The negative mean logrank.

    Raises:
        ValueError: If the shapes of `labels` and `logits` are incompatible or batch size is > 1.
    """
    validate_tensor_shapes(labels, logits)

    matches: torch.Tensor = (
        logits.argsort(-1, descending=True) == labels.unsqueeze(-1)
    ).nonzero()
    ranks: torch.Tensor = matches[:, -1]

    log_ranks: torch.Tensor = torch.log(ranks.float() + 1)

    return -log_ranks.mean().item()


def likelihood_logrank_ratio(labels: torch.Tensor, logits: torch.Tensor) -> float:
    """Compute the Likelihood Logrank Ratio (LRR) from DetectLLM paper.

    Args:
        labels (torch.Tensor): Ground truth labels of shape: (1, sequence_length).
        logits (torch.Tensor): Logits of shape: (1, sequence_length, vocab_size).

    Returns:
        float: The LRR Ratio.

    Raises:
        ValueError: If the shapes of `labels` and `logits` are incompatible or batch size is > 1.
    """
    validate_tensor_shapes(labels, logits)

    _log_likelihood: float = log_likelihood(labels, logits)
    _log_rank: float = log_rank(labels, logits)

    return _log_likelihood / _log_rank


def fast_detect_gpt(labels: torch.Tensor, logits: torch.Tensor) -> float:
    """Estimate the conditional probability gap using FastDetectGPT.

    Args:
        labels (torch.Tensor): Ground truth labels of shape: (1, sequence_length).
        logits (torch.Tensor): Logits of shape: (1, sequence_length, vocab_size).

    Returns:
        float: The estimated mean gap which serves as the detection metric.

    Raises:
        ValueError: If the shapes of `labels` and `logits` are incompatible or batch size is > 1.
    """
    validate_tensor_shapes(labels, logits)

    # conditional sampling
    log_probs: torch.Tensor = F.log_softmax(logits, dim=-1)
    distribution: torch.distributions.categorical.Categorical = (
        torch.distributions.categorical.Categorical(logits=log_probs)
    )
    x_tilde: torch.Tensor = distribution.sample([10000]).permute([1, 2, 0])

    log_likelihood_x: torch.Tensor = log_probs.gather(
        dim=-1, index=labels.unsqueeze(-1)
    ).mean(dim=1)
    log_likelihood_x_tilde: torch.Tensor = log_probs.gather(dim=-1, index=x_tilde).mean(
        dim=1
    )

    # estimate the mean/variance
    mu_tilde: torch.Tensor = log_likelihood_x_tilde.mean(dim=-1)
    sigma_tilde: torch.Tensor = log_likelihood_x_tilde.std(dim=-1)

    # estimate conditional probability curvature
    dhat: torch.Tensor = (log_likelihood_x - mu_tilde) / sigma_tilde

    return dhat.item()



import torch
import pytest
from torch import Tensor

from pydetectgpt import (
    log_likelihood,
    log_rank,
    likelihood_logrank_ratio,
    fast_detect_gpt,
)


def test_log_likelihood():
    # shape mismatch
    logits: Tensor = torch.randn(1, 5, 10)
    labels: Tensor = torch.randint(0, 9, (1, 6))

    with pytest.raises(
        ValueError, match="Labels and logits must have compatible shapes"
    ):
        log_likelihood(labels, logits)

    # batch size > 1
    logits = torch.randn(2, 5, 10)
    labels = torch.randint(0, 9, (2, 5))

    with pytest.raises(ValueError, match="Batch size must be 1"):
        log_likelihood(labels, logits)

    # label > vocab size
    logits = torch.randn(1, 3, 10)
    labels = torch.tensor([[2, 5, 10]])

    with pytest.raises(ValueError, match="Labels must be in vocab size"):
        log_likelihood(labels, logits)

    # some simple tests I calculated manually
    logits = torch.tensor([[[0.2, 0.3, 0.4]]])
    labels = torch.tensor([[1]])

    assert abs(log_likelihood(labels, logits) - -1.1019428) < 1e-5

    logits = torch.tensor([[[2.3, 1.1, 0.5], [0.8, 2.5, 1.1], [1.5, 2.1, 0.2]]])
    labels = torch.tensor([[0, 1, 2]])

    assert abs(log_likelihood(labels, logits) - -1.05657327) < 1e-5


def test_log_rank():
    # shape mismatch
    logits: Tensor = torch.randn(1, 5, 10)
    labels: Tensor = torch.randint(0, 9, (1, 6))

    with pytest.raises(
        ValueError, match="Labels and logits must have compatible shapes"
    ):
        log_rank(labels, logits)

    # batch size > 1
    logits = torch.randn(2, 5, 10)
    labels = torch.randint(0, 9, (2, 5))

    with pytest.raises(ValueError, match="Batch size must be 1"):
        log_rank(labels, logits)

    # label > vocab size
    logits = torch.randn(1, 3, 10)
    labels = torch.tensor([[2, 5, 10]])

    with pytest.raises(ValueError, match="Labels must be in vocab size"):
        log_rank(labels, logits)

    # some simple tests I calculated manually
    logits = torch.tensor([[[0.2, 0.3, 0.4]]])
    labels = torch.tensor([[1]])

    assert abs(log_rank(labels, logits) - -0.693147) < 1e-5

    logits = torch.tensor([[[2.3, 1.1, 0.5], [0.8, 2.5, 1.1], [1.5, 2.1, 0.2]]])
    labels = torch.tensor([[0, 1, 2]])

    assert abs(log_rank(labels, logits) - -0.3662) < 1e-5


def test_likelihood_logrank_ratio():
    # shape mismatch
    logits: Tensor = torch.randn(1, 5, 10)
    labels: Tensor = torch.randint(0, 9, (1, 6))

    with pytest.raises(
        ValueError, match="Labels and logits must have compatible shapes"
    ):
        likelihood_logrank_ratio(labels, logits)

    # batch size > 1
    logits = torch.randn(2, 5, 10)
    labels = torch.randint(0, 9, (2, 5))

    with pytest.raises(ValueError, match="Batch size must be 1"):
        likelihood_logrank_ratio(labels, logits)

    # label > vocab size
    logits = torch.randn(1, 3, 10)
    labels = torch.tensor([[2, 5, 10]])

    with pytest.raises(ValueError, match="Labels must be in vocab size"):
        likelihood_logrank_ratio(labels, logits)

    # some simple tests I calculated manually
    logits = torch.tensor([[[0.2, 0.3, 0.4]]])
    labels = torch.tensor([[1]])

    assert abs(likelihood_logrank_ratio(labels, logits) - 1.5897675) < 1e-5

    logits = torch.tensor([[[2.3, 1.1, 0.5], [0.8, 2.5, 1.1], [1.5, 2.1, 0.2]]])
    labels = torch.tensor([[0, 1, 2]])

    assert abs(likelihood_logrank_ratio(labels, logits) - 2.8852) < 1e-5


def test_fast_detect_gpt():
    # shape mismatch
    logits: Tensor = torch.randn(1, 5, 10)
    labels: Tensor = torch.randint(0, 9, (1, 6))

    with pytest.raises(
        ValueError, match="Labels and logits must have compatible shapes"
    ):
        fast_detect_gpt(labels, logits)

    # batch size > 1
    logits = torch.randn(2, 5, 10)
    labels = torch.randint(0, 9, (2, 5))

    with pytest.raises(ValueError, match="Batch size must be 1"):
        fast_detect_gpt(labels, logits)

    # label > vocab size
    logits = torch.randn(1, 3, 10)
    labels = torch.tensor([[2, 5, 10]])

    with pytest.raises(ValueError, match="Labels must be in vocab size"):
        fast_detect_gpt(labels, logits)

    # some simple tests I calculated manually
    # since sampling is used in fastdetectgpt I lowered the difference threshold
    logits = torch.tensor([[[0.2, 0.3, 0.4]]])
    labels = torch.tensor([[1]])

    assert abs(fast_detect_gpt(labels, logits) - -0.085) < 5e-2

    logits = torch.tensor([[[2.3, 1.1, 0.5], [0.8, 2.5, 1.1], [1.5, 2.1, 0.2]]])
    labels = torch.tensor([[0, 1, 2]])

    assert abs(fast_detect_gpt(labels, logits) - -0.55) < 5e-2



"""On device LLM-Generated text detection in Pytorch."""

from .detect import detect_ai_text
from .methods import log_likelihood, log_rank, likelihood_logrank_ratio, fast_detect_gpt

__version__ = "0.1.0"
__all__ = [
    "detect_ai_text",
    "log_likelihood",
    "log_rank",
    "likelihood_logrank_ratio",
    "fast_detect_gpt",
]



"""CLI for using PyDetectGPT."""

import argparse
from .detect import detect_ai_text, DETECTION_FUNCS, THRESHOLDS


def main() -> None:
    """Run detection via CLI with argparse."""
    parser = argparse.ArgumentParser(
        description="Detect if text is AI-generated",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""
Example:
  pydetectgpt "Your text here"
""",
    )

    parser.add_argument("text", help="text to analyze")

    parser.add_argument(
        "-m",
        "--method",
        type=str,
        choices=list(DETECTION_FUNCS.keys()),
        default="fastdetectgpt",
        help="detection method to use (default: %(default)s)",
    )

    parser.add_argument(
        "--model",
        type=str,
        default="Qwen/Qwen2.5-1.5B",
        help="HuggingFace model to use (default: %(default)s)",
    )

    parser.add_argument(
        "-t",
        "--threshold",
        type=float,
        help=f"decision threshold (defaults: {THRESHOLDS})",
    )

    parser.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        help="output only the result (0 for human, 1 for AI)",
    )

    args = parser.parse_args()

    result = detect_ai_text(
        text=args.text,
        method=args.method,
        detection_model=args.model,
        threshold=args.threshold,
    )

    if args.quiet:
        print(result)
    else:
        print(f"Detection Result: {'AI Generated' if result else 'Human Written'}")


if __name__ == "__main__":
    main()



import pytest
from unittest.mock import patch
from pydetectgpt.cli import main

AI_TEXT: str = (
    "Texas A&M University is located in College Station, Texas, in the southeastern part of the state. It's about 90 miles northwest of Houston and around 150 miles south of Dallas. The university's full name is Texas Agricultural and Mechanical University, and it is one of the largest public universities in the United States."
)
HUMAN_TEXT: str = (
    "The main problem the authors are trying to address is that Large Language Models require large computational resources to use. This means that as a common setup we see companies deploying GPU clusters which act as a cloud server to generate responses when a user presents a query. Aside from the vast resources needed to set up a GPU cluster this approach has 2 main downsides: sending queries over the internet via an API exposes users' private data and results in additional latency when generating responses"
)


def test_cli_loglikelihood(capsys) -> None:
    method: str = "loglikelihood"

    with patch("sys.argv", ["pydetectgpt", AI_TEXT, "-m", method]):
        main()
        captured = capsys.readouterr()
        assert "AI Generated" in captured.out

    with patch("sys.argv", ["pydetectgpt", HUMAN_TEXT, "-m", method]):
        main()
        captured = capsys.readouterr()
        assert "Human Written" in captured.out

    with patch("sys.argv", ["pydetectgpt", AI_TEXT, "-m", method, "-t", "99999.9"]):
        main()
        captured = capsys.readouterr()
        assert "Human Written" in captured.out

    with patch("sys.argv", ["pydetectgpt", HUMAN_TEXT, "-m", method, "-t", "-99999.9"]):
        main()
        captured = capsys.readouterr()
        assert "AI Generated" in captured.out


def test_cli_logrank(capsys) -> None:
    method: str = "logrank"

    with patch("sys.argv", ["pydetectgpt", AI_TEXT, "-m", method]):
        main()
        captured = capsys.readouterr()
        assert "AI Generated" in captured.out

    with patch("sys.argv", ["pydetectgpt", HUMAN_TEXT, "-m", method]):
        main()
        captured = capsys.readouterr()
        assert "Human Written" in captured.out

    with patch("sys.argv", ["pydetectgpt", AI_TEXT, "-m", method, "-t", "99999.9"]):
        main()
        captured = capsys.readouterr()
        assert "Human Written" in captured.out

    with patch("sys.argv", ["pydetectgpt", HUMAN_TEXT, "-m", method, "-t", "-99999.9"]):
        main()
        captured = capsys.readouterr()
        assert "AI Generated" in captured.out


def test_cli_detectllm(capsys) -> None:
    """Test CLI with detectllm method."""
    method: str = "detectllm"

    with patch("sys.argv", ["pydetectgpt", AI_TEXT, "-m", method]):
        main()
        captured = capsys.readouterr()
        assert "AI Generated" in captured.out

    with patch("sys.argv", ["pydetectgpt", HUMAN_TEXT, "-m", method]):
        main()
        captured = capsys.readouterr()
        assert "Human Written" in captured.out

    with patch("sys.argv", ["pydetectgpt", AI_TEXT, "-m", method, "-t", "99999.9"]):
        main()
        captured = capsys.readouterr()
        assert "Human Written" in captured.out

    with patch("sys.argv", ["pydetectgpt", HUMAN_TEXT, "-m", method, "-t", "-99999.9"]):
        main()
        captured = capsys.readouterr()
        assert "AI Generated" in captured.out


def test_cli_fastdetectgpt(capsys) -> None:
    method: str = "fastdetectgpt"

    with patch("sys.argv", ["pydetectgpt", AI_TEXT, "-m", method]):
        main()
        captured = capsys.readouterr()
        assert "AI Generated" in captured.out

    with patch("sys.argv", ["pydetectgpt", HUMAN_TEXT, "-m", method]):
        main()
        captured = capsys.readouterr()
        assert "Human Written" in captured.out

    with patch("sys.argv", ["pydetectgpt", AI_TEXT, "-m", method, "-t", "99999.9"]):
        main()
        captured = capsys.readouterr()
        assert "Human Written" in captured.out

    with patch("sys.argv", ["pydetectgpt", HUMAN_TEXT, "-m", method, "-t", "-99999.9"]):
        main()
        captured = capsys.readouterr()
        assert "AI Generated" in captured.out


def test_cli_invalid_method() -> None:
    with patch("sys.argv", ["pydetectgpt", AI_TEXT, "-m", "invalid_method"]):
        with pytest.raises(SystemExit):
            main()


def test_cli_quiet_mode(capsys) -> None:
    with patch("sys.argv", ["pydetectgpt", AI_TEXT, "-q"]):
        main()
        captured = capsys.readouterr()
        assert captured.out.strip() == "1"

    with patch("sys.argv", ["pydetectgpt", HUMAN_TEXT, "-q"]):
        main()
        captured = capsys.readouterr()
        assert captured.out.strip() == "0"



from pydetectgpt import detect_ai_text
import pytest
from pydetectgpt.detect import DetectionMethod

# I asked chatgpt "Where is Texas A&M?"
AI_TEXT: str = (
    "Texas A&M University is located in College Station, Texas, in the southeastern part of the state. It's about 90 miles northwest of Houston and around 150 miles south of Dallas. The university's full name is Texas Agricultural and Mechanical University, and it is one of the largest public universities in the United States."
)
# random paragraph from one of my assignments (written by human)
HUMAN_TEXT: str = (
    "The main problem the authors are trying to address is that Large Language Models require large computational resources to use. This means that as a common setup we see companies deploying GPU clusters which act as a cloud server to generate responses when a user presents a query. Aside from the vast resources needed to set up a GPU cluster this approach has 2 main downsides: sending queries over the internet via an API exposes users’ private data and results in additional latency when generating responses"
)


def test_detect_ai_text():
    # invalid method name
    with pytest.raises(ValueError, match="must be one of"):
        detect_ai_text(AI_TEXT, method="notvalidmethodname")

    # empty text should be human
    assert detect_ai_text("") == 0

    # high threshold should always be 0 (human)
    assert detect_ai_text(AI_TEXT, threshold=99999.9) == 0

    # low threshold should always be 1 (ai)
    assert detect_ai_text(HUMAN_TEXT, threshold=-99999.9) == 1


def test_detect_ai_text_loglikelihood():
    method: DetectionMethod = "loglikelihood"

    assert detect_ai_text(AI_TEXT, method=method) == 1

    assert detect_ai_text(HUMAN_TEXT, method=method) == 0

    assert detect_ai_text(AI_TEXT, method=method, threshold=99999.9) == 0

    assert detect_ai_text(HUMAN_TEXT, method=method, threshold=-99999.9) == 1


def test_detect_ai_text_logrank():
    method: DetectionMethod = "logrank"

    assert detect_ai_text(AI_TEXT, method=method) == 1

    assert detect_ai_text(HUMAN_TEXT, method=method) == 0

    assert detect_ai_text(AI_TEXT, method=method, threshold=99999.9) == 0

    assert detect_ai_text(HUMAN_TEXT, method=method, threshold=-99999.9) == 1


def test_detect_ai_text_detectllm():
    method: DetectionMethod = "detectllm"

    assert detect_ai_text(AI_TEXT, method=method) == 1

    assert detect_ai_text(HUMAN_TEXT, method=method) == 0

    assert detect_ai_text(AI_TEXT, method=method, threshold=99999.9) == 0

    assert detect_ai_text(HUMAN_TEXT, method=method, threshold=-99999.9) == 1


def test_detect_ai_text_fastdetectgpt():
    method: DetectionMethod = "fastdetectgpt"

    assert detect_ai_text(AI_TEXT, method=method) == 1

    assert detect_ai_text(HUMAN_TEXT, method=method) == 0

    assert detect_ai_text(AI_TEXT, method=method, threshold=99999.9) == 0

    assert detect_ai_text(HUMAN_TEXT, method=method, threshold=-99999.9) == 1



"""Utils used throughout source code."""

import torch
from transformers import AutoTokenizer, AutoModelForCausalLM
from typing import Tuple


def load_model(hf_repo: str) -> Tuple[AutoModelForCausalLM, AutoTokenizer]:
    """Load model and tokenizer from the Hugging Face repository.

    Args:
        hf_repo (str): The Hugging Face model repository identifier (e.g., 'Qwen/Qwen2.5-1.5B').

    Returns:
        Tuple[AutoModelForCausalLM, AutoTokenizer]: A tuple containing the model and tokenizer.

    Raises:
        ValueError: If there is an issue loading the model or tokenizer from HuggingFace.
    """
    device: str = "cuda" if torch.cuda.is_available() else "cpu"
    tokenizer: AutoTokenizer = AutoTokenizer.from_pretrained(hf_repo)
    model: AutoModelForCausalLM = AutoModelForCausalLM.from_pretrained(hf_repo).to(
        device
    )

    if tokenizer.pad_token_id is None:
        tokenizer.pad_token_id = tokenizer.eos_token_id

    return model, tokenizer


def validate_tensor_shapes(labels: torch.Tensor, logits: torch.Tensor) -> None:
    """Validate the compatibility of labels and logits.

    Args:
        labels (torch.Tensor): Ground truth labels of shape: (1, sequence_length).
        logits (torch.Tensor): Logits of shape: (1, sequence_length, vocab_size).

    Raises:
        ValueError: If the shapes of `labels` and `logits` are incompatible or batch size is > 1.
    """
    if logits.shape[0] != 1 or labels.shape[0] != 1:
        raise ValueError(
            f"Batch size must be 1, but got logits batch size {logits.shape[0]} "
            f"and labels batch size {labels.shape[0]}"
        )

    if logits.dim() < 2:
        raise ValueError(
            f"Logits must have at least 2 dimensions, but got shape {logits.shape}"
        )

    if labels.shape != logits.shape[:-1]:
        raise ValueError(
            f"Labels and logits must have compatible shapes. "
            f"Got labels shape {labels.shape} and logits shape {logits.shape[:-1]}"
        )

    if labels.max().item() >= logits.shape[-1]:
        raise ValueError(
            f"Labels must be in vocab size ({logits.shape[-1]}), "
            f"but got label {labels.max().item()}"
        )
